/* File.....: dsu_udfs.prg
 * Author...: Peter Townsend
 * CIS ID...:
 * e-mail...: cephas@tpgi.com.au
 *
 * A set of functions to save and restore database information.
 *     DSU_Save()
 *     DSU_Rest()
 *     OrdBagsGet()
 *     OrdBagsSet()
 *     RelsGet()
 *     RelsSet()
 *
 * Assumptions: 1. That all databases are open in contiguous areas.
 *                 (that is, that all areas between the first and the
 *                  last area are used.)
 *              2. That the first area used is area 1.
 *              3. That no databases are opened read-only.
 *              4. That a database name is the same as its alias.
 *
 * NB: This function depends on my SaveArry() and LoadArry() functions
 *     which can save and restore multi-level arrays.
 *     It also uses my version of Victor Szol's OpenMode() function.
 */
/*
 * This is an original work by Peter Townsend and is
 * hereby placed in the public domain.
 */
#include "common.ch"
#include "openmode.ch"
#define DSU_FILE "DSETUP.ARR"
#define DSU_DBF 1    && DBF name                    within DBF info
#define DSU_RDD 2    && Replaceable Database driver within DBF info
#define DSU_MOD 3    && Open mode                   within DBF info
#define DSU_REC 4    && Current record no           within DBF info
#define DSU_LOK 5    && Records locked              within DBF info
#define DSU_FLT 6    && Filter in use               within DBF info
#define DSU_REL 7    && Relations                   within DBF info
#define DSU_BAG 8    && Index names                 within DBF info
#define DSU_ORD 9    && Controlling index           within DBF info
#define DSU_QTY 9    && No of items maintained in DBF info
/*  $DOC$
 *  $FUNCNAME$
 *      DSU_Save()
 *
 *  $ONELINER$
 *      Save all information about open databases to a file.
 *
 *  $SYNTAX$
 *      DSU_Save( [<cFileName>] ) -> lSuccess
 *
 *  $ARGUMENTS$
 *      cFileName  - name of file into which information is to be stored
 *                   (Default value for cFileName is "DSETUP.ARR")
 *
 *  $RETURNS$
 *      lSuccess   - was the information successfully saved?
 *
 *  $DESCRIPTION$
 *      This function saves a list of all open databases to a file.
 *      The following information is saved for each open database:
 *          1. The name (alias) of the database file
 *          2. The RDD used to access the database file
 *          3. Whether the database file is open for sharing
 *          4. The current record number
 *          5. The list of record locks on this database
 *          6. The current filter condition
 *          7. The relations set from this database
 *          8. The index files which are open
 *          9. The controlling index
 *
 *  $EXAMPLES$
 *      IF DSU_Save()
 *          DBCLOSEALL()
 *          .
 *          * Housekeeping
 *          .
 *          IF ! DSU_Rest()
 *              ? "Cannot restore file setup"
 *              StdOpen()    // Open standard file list
 *          ENDIF
 *       ENDIF
 *
 *  $SEEALSO$
 *      DSU_Rest()
 *
 *  $INCLUDE$
 *      common.ch
 *      openmode.ch
 *
 *  $END$
 */
FUNCTION DSU_Save(cFile)
    LOCAL nDbfCntr, aDBInfo, nCrntArea, aDBUnit
    DEFAULT cFile TO DSU_FILE
    *
    * Save environment
    *
    nCrntArea := SELECT()
    *
    * Load database information for each used area.
    *
    aDBInfo   := {}
    nDbfCntr  := 1
    DBSELECTAREA(nDbfCntr)
    DO WHILE USED()
        aDBUnit := ARRAY(DSU_QTY)
        aDBUnit[DSU_DBF] := ALIAS()
        aDBUnit[DSU_RDD] := RDDNAME()
        aDBUnit[DSU_MOD] := (OpenMode(ALIAS()) == OM_SHARED)
        aDBUnit[DSU_REC] := RECNO()
        aDBUnit[DSU_LOK] := DBRLOCKLIST()
        aDBUnit[DSU_FLT] := DBFILTER()
        aDBUnit[DSU_REL] := RelsGet()
        aDBUnit[DSU_BAG] := OrdBagsGet()
        aDBUnit[DSU_ORD] := INDEXORD()
        AADD(aDBInfo, aDBUnit)
        nDbfCntr++
        DBSELECTAREA(nDbfCntr)
    ENDDO
    *
    * Save array to disk
    *
    SaveArry(aDBInfo, cFile)
    *
    * Restore environment
    *
    DBSELECTAREA(nCrntArea)
RETURN(NIL)
/*  $DOC$
 *  $FUNCNAME$
 *      DSU_Rest()
 *
 *  $ONELINER$
 *      Load all information about databases/indexes to open from a file.
 *
 *  $SYNTAX$
 *      DSU_Rest( [<cFileName>] ) -> lSuccess
 *
 *  $ARGUMENTS$
 *      cFileName  - name of file from which information is to be restored
 *                   (Default value for cFileName is "DSETUP.ARR")
 *
 *  $RETURNS$
 *      lSuccess   - was the information successfully restored?
 *
 *  $DESCRIPTION$
 *      This function loads a list of all databases to be opened from
 *      a file.  Associated information which is also loaded includes
 *      the current record no, the RDD, whether the database is opened
 *      SHARED, the list of index files, and the order of the controlling
 *      index.  This information is used to set the databases up in
 *      exactly the manner of their use at the time that the information
 *      was saved.
 *
 *  $EXAMPLES$
 *      IF DSU_Save()
 *          DBCLOSEALL()
 *          .
 *          * Housekeeping
 *          .
 *          IF !DSU_Rest()
 *              ? "Cannot restore file setup"
 *              StdOpen()    // Open standard file list
 *          ENDIF
 *      ENDIF
 *
 *  $SEEALSO$
 *      DSU_Save()
 *
 *  $INCLUDE$
 *      common.ch
 *
 *  $END$
 */
FUNCTION DSU_Rest(cFile)
    LOCAL aDbfList, nDbfCntr, aDBUnit, lSuccess, aLokList, aRelations
    DEFAULT cFile TO DSU_FILE
    lSuccess := .T.
    aDbfList := {}
    IF FILE(cFile)
        aDbfList := LoadArry(cFile)
        IF aDbfList == NIL
            lSuccess := .F.
        ENDIF
    ELSE
        lSuccess := .F.
    ENDIF
    IF lSuccess
        DBCLOSEALL()
        aRelations := ARRAY(LEN(aDbfList))
        *
        * Restore database files
        *
        FOR nDbfCntr := 1 TO LEN(aDbfList)
            aDBUnit := aDbfList[nDbfCntr]
            DBSELECTAREA(nDbfCntr)
            DBUSEAREA(.F., aDBUnit[DSU_RDD], aDBUnit[DSU_DBF], ;
                , aDBUnit[DSU_MOD])
            IF USED()
                *
                * Restore indexes
                *
                IF OrdBagsSet(aDBUnit[DSU_BAG])
                    DBSETORDER(aDBUnit[DSU_ORD])
                ELSE
                    lSuccess := .F.
                ENDIF
                *
                * Re-lock locked records.
                *
                aLokList := aDBUnit[DSU_LOK]
                FOR nLokCntr := 1 TO LEN(aLokList)
                    IF ! DBRLOCK(aLokList[nLokCntr])
                        lSuccess := .F.
                    ENDIF
                NEXT
                *
                * Re-setablish filter condition.
                *
                SET FILTER TO &(aDBUnit[DSU_FLT])
                *
                * Restore record position.
                *
                DBGOTO(aDBUnit[DSU_REC])
                *
                * Save relationship info to array for setting later.
                *
                aRelations[nDbfCntr] := {SELECT(), aDBUnit[DSU_REL]}
            ELSE
                ? "Cannot open database " + aDBUnit[DSU_DBF]
                lSuccess := .F.
            ENDIF
        NEXT
        *
        * If all database/index opens successful, set relationships.
        *
        IF lSuccess
            FOR nDbfCntr := 1 TO LEN(aRelations)
                IF ! EMPTY(aRelations[nDbfCntr])
                    SELECT (aRelations[nDbfCntr][1])
                    IF ! RelsSet(aRelations[nDbfCntr][2])
                        lSuccess := .F.
                    ENDIF
                ENDIF
            NEXT
        ENDIF
    ENDIF
RETURN(lSuccess)
/*  $DOC$
 *  $FUNCNAME$
 *      OrdBagsGet()
 *
 *  $ONELINER$
 *      Returns an array of the open indexes for the current area.
 *
 *  $SYNTAX$
 *      OrdBagsGet() -> aOrdBags
 *
 *  $ARGUMENTS$
 *
 *  $RETURNS$
 *      An array of the index files in use for the current area.
 *
 *  $DESCRIPTION$
 *      This function obtains a list of the current open index files
 *      for the current area, returning the list in an array of index names.
 *
 *  $EXAMPLES$
 *      aCustBags := customer->(OrdBagsGet())
 *      customer->(DBCLOSEAREA())
 *      .
 *      .
 *      SELECT 0
 *      USE customer
 *      IF ! OrdBagsSet(aCustBags)
 *          ? "Error re-opening indexes"
 *          QUIT()
 *      ENDIF
 *
 *  $SEEALSO$
 *      OrdBagsSet()
 *
 *  $INCLUDE$
 *
 *  $END$
 */
FUNCTION OrdBagsGet()
    LOCAL aBagList, nBagCntr
    aBagList := {}
    nBagCntr := 1
    DO WHILE ! EMPTY(ORDBAGNAME(nBagCntr))
        AADD(aBagList, ORDBAGNAME(nBagCntr))
        nBagCntr++
    ENDDO
RETURN(aBagList)
/*  $DOC$
 *  $FUNCNAME$
 *      OrdBagsSet()
 *
 *  $ONELINER$
 *      For the current area, open the indexes passed in as an array.
 *
 *  $SYNTAX$
 *      OrdBagsSet( <aOrdBags> ) -> lSuccess
 *
 *  $ARGUMENTS$
 *      aOrdBags - an array of indexes, obtained from OrdBagsGet()
 *
 *  $RETURNS$
 *      lSuccess - were the index openings successful?
 *
 *  $DESCRIPTION$
 *      This function closes all open indexes for the current area,
 *      then opens the indexes passed in as an array.
 *
 *  $EXAMPLES$
 *      aCustBags := customer->(OrdBagsGet())
 *      customer->(DBCLOSEAREA())
 *      .
 *      .
 *      SELECT 0
 *      USE customer
 *      IF ! OrdBagsSet(aCustBags)
 *          ? "Error re-opening indexes"
 *          QUIT()
 *      ENDIF
 *
 *  $SEEALSO$
 *      OrdBagsGet()
 *
 *  $INCLUDE$
 *
 *  $END$
 */
FUNCTION OrdBagsSet(aBagList)
    LOCAL nBagCntr, lSuccess
    lSuccess := .T.
    ORDLISTCLEAR()
    FOR nBagCntr := 1 TO LEN(aBagList)
        ORDLISTADD(aBagList[nBagCntr])
        IF ORDBAGNAME(nBagCntr) != aBagList[nBagCntr]
            lSuccess := .F.
        ENDIF
    NEXT
RETURN(lSuccess)
/*  $DOC$
 *  $FUNCNAME$
 *      RelsGet()
 *
 *  $ONELINER$
 *      Returns an array of the relations set for the current area.
 *
 *  $SYNTAX$
 *      RelsGet() -> aRelations
 *
 *  $ARGUMENTS$
 *
 *  $RETURNS$
 *      aRelations - An array of the relations set for the current area.
 *
 *  $DESCRIPTION$
 *      This function obtains a list of the relations for the current area,
 *      returning the list in an array.
 *
 *  $EXAMPLES$
 *      aCustRels := customer->(RelsGet())
 *      DBCLEARRELATION()
 *      .
 *      .
 *      IF ! RelsSet(aCustRels)
 *          ? "Error re-setting relations"
 *          QUIT()
 *      ENDIF
 *
 *  $SEEALSO$
 *      RelsSet()
 *
 *  $INCLUDE$
 *
 *  $END$
 */
FUNCTION RelsGet()
    LOCAL aRelList, nRelCntr
    aRelList := {}
    nRelCntr := 1
    DO WHILE ! EMPTY(DBRELATION(nRelCntr))
        AADD(aRelList, {DBRSELECT(nRelCntr), DBRELATION(nRelCntr)})
        nRelCntr++
    ENDDO
RETURN(aRelList)
/*  $DOC$
 *  $FUNCNAME$
 *      RelsSet()
 *
 *  $ONELINER$
 *      For the current area, set the relations passed in as an array.
 *
 *  $SYNTAX$
 *      RelsSet( <aRelations> ) -> lSuccess
 *
 *  $ARGUMENTS$
 *      aRelations - an array of relations, obtained from RelsGet()
 *
 *  $RETURNS$
 *      lSuccess - were the relations successfully set?
 *
 *  $DESCRIPTION$
 *      This function clears all relations for the current area,
 *      then sets the relations passed in as an array.
 *
 *  $EXAMPLES$
 *      aCustRels := customer->(RelsGet())
 *      customer->(DBCLEARRELATION())
 *      .
 *      .
 *      lRetry := .T.
 *      DO WHILE lRetry
 *          lRetry := .F.
 *          IF ! customer->(RelsSet(aCustRels))
 *              IF ALERT("Error re-setting relations - retry?", {"yes", "no"}) == 1
 *                  lRetry := .T.
 *              ENDIF
 *          ENDIF
 *      ENDDO
 *
 *  $SEEALSO$
 *      RelsGet()
 *
 *  $INCLUDE$
 *
 *  $END$
 */
FUNCTION RelsSet(aRelList)
    LOCAL nRelCntr, lSuccess
    lSuccess := .T.
    DBCLEARRELATION()
    FOR nRelCntr := 1 TO LEN(aRelList)
        IF EMPTY(ALIAS()) .OR. EMPTY(ALIAS(aRelList[nRelCntr][1]))
            * The pair of databases necessary aren't open.
            lSuccess := .F.
        ELSE
            SET RELATION TO &(aRelList[nRelCntr][2]) ;
                INTO &(ALIAS(aRelList[nRelCntr][1])) ADDITIVE
            IF EMPTY(DBRELATION(nRelCntr))
                * The relation wasn't set.
                lSuccess := .F.
            ENDIF
        ENDIF
    NEXT
RETURN(lSuccess)

