* ######################################################################### *
* These functions store an array to disk, and will load an array from disk. *
* They use recursion to perform both operations in order to keep the code   *
* size as small as possible and to keep the efficiency as high as possible. *
* Note that the elements do not have to be all of the same type, and that   *
* even arrays which are non-uniformly populated are still processed.        *
* The only limitation (as can be seen from the defines below) is that blocks*
* and objects cannot be processed and so are replaced by a text message to  *
* that effect.                                                              *
* Author: Peter Townsend   Date: 11-12-1996                                 *
* ######################################################################### *
#include "fileio.ch"
#include "common.ch"
#define TYPELEN 1               // No. chars for variable TYPE
#define VALUELEN 5              // No. bytes for variable LENGTH
#define NELEMLEN 6              // No. bytes for NO. OF ELEMENTS in array
#define DATELEN 8               // No. bytes for a DATE variable
#define LOGICALLEN 1            // No. bytes for a LOGIC variable
#define BLOCKMSG "CANNOT STORE BLOCK"   // Used in place of the BLOCK type
#define OBJECTMSG "CANNOT STORE OBJECT" // Used in place of the OBJECT type
* ========================================================================= *
* SaveArry  - Save an array to disk.                                        *
*                                                                           *
* Parameters: <ExpA1> - the array to store                                  *
*             <ExpC1> - name of file to store array in                      *
*                       (optional, defaults to "STORED.ARR")                *
*                                                                           *
* Returns   : <ExpL>  - was the file created?                               *
*                                                                           *
* Example:   MyArray := {1, 2, 3, 4}                                        *
*            IF SaveArry(MyArray, "myarray.arr")                            *
*                ? "Array stored"                                           *
*            ENDIF                                                          *
* ========================================================================= *
FUNCTION SaveArry(aArray, cFile)
    LOCAL nHandle, lAbort
    DEFAULT cFile TO "STORED.ARR"
    lAbort := .F.
    IF VALTYPE(aArray) $ "CDLN"
        aArray := {aArray}
    ENDIF
    IF VALTYPE(aArray) != "A"
        lAbort := .T.
    ENDIF
    IF !lAbort
        nHandle := FCREATE(cFile)
        SaveLevel(aArray, nHandle)
        FCLOSE(nHandle)
    ENDIF
    IF lAbort
        IF FILE(cFile)
            DELETE FILE (cFile)
        ENDIF
    ENDIF
RETURN(FILE(cFile))
* ========================================================================= *
* SaveLevel  - Save 1 dimension of the array, recursively processing any    *
*              arrays which may be part of that dimension.                  *
*                                                                           *
* Parameters: <ExpA1> - the dimension to save.                              *
*             <ExpN1> - the handle for the file in which array is to be     *
*                       saved.                                              *
*                                                                           *
* Returns   : NIL                                                           *
* ========================================================================= *
STATIC FUNCTION SaveLevel(pArray, nHandle)
    LOCAL cType, nElems, nElemCntr, nNumLen, cStr
    nElems := LEN(pArray)
    FWRITE(nHandle, ntosz(nElems, NELEMLEN))
    FOR nElemCntr := 1 TO nElems
        cType  := VALTYPE(pArray[nElemCntr])
        FWRITE(nHandle, cType)
        DO CASE
            CASE cType == "A"
                * Array
                SaveLevel(pArray[nElemCntr], nHandle)
            CASE cType == "B"
                * Block
                FWRITE(nHandle, ntosz(LEN(BLOCKMSG), VALUELEN))
                FWRITE(nHandle, BLOCKMSG)
            CASE cType == "C"
                * Character
                FWRITE(nHandle, ntosz(LEN(pArray[nElemCntr]), VALUELEN))
                FWRITE(nHandle, pArray[nElemCntr])
            CASE cType == "D"
                * Date
                FWRITE(nHandle, ntosz(DATELEN, VALUELEN))
                FWRITE(nHandle, DTOS(pArray[nElemCntr]))
            CASE cType == "L"
                * Logical
                FWRITE(nHandle, ntosz(LOGICALLEN, VALUELEN))
                FWRITE(nHandle, IF(pArray[nElemCntr], "T", "F"))
            CASE cType == "M"
                * Memo (should never be used)
                FWRITE(nHandle, ntosz(LEN(pArray[nElemCntr]), VALUELEN))
                FWRITE(nHandle, pArray[nElemCntr])
            CASE cType == "N"
                * Number
                cStr    := ALLTRIM(TRANSFORM(pArray[nElemCntr], "@B"))
                nNumLen := LEN(cStr)
                FWRITE(nHandle, ntosz(nNumLen, VALUELEN))
                FWRITE(nHandle, cStr)
            CASE cType == "O"
                * Object
                FWRITE(nHandle, ntosz(LEN(OBJECTMSG), VALUELEN))
                FWRITE(nHandle, OBJECTMSG)
        ENDCASE
    NEXT
RETURN(NIL)
* ========================================================================= *
* LoadArry  - Load an array from disk.                                      *
*                                                                           *
* Parameters: <ExpC1> - the name of the file to be read.                    *
*                                                                           *
* Returns   : <ExpA>  - the array which was stored in the file, if          *
*                       successful, or NIL if not successful.               *
*                                                                           *
* Example:   MyArray := LoadArry("myarray.arr")                             *
*            IF MyArray != NIL                                              *
*                ? "Array retrieved"                                        *
*            ENDIF                                                          *
* ========================================================================= *
FUNCTION LoadArry(cFile)
    LOCAL nHandle, lOK, aRetVal
    DEFAULT cFile TO "STORED.ARR"
    lOK     := .F.
    aRetVal := NIL   // Unnecessary, but sets a value which may be used.
    nHandle := -1    // Set to the error value for an open.
    IF FILE(cFile)
        nHandle := FOPEN(cFile, FO_READ)
        lOK     := (nHandle > -1)
    ENDIF
    IF lOK
        aRetVal := LoadLevel(nHandle)  // Recursively process the array.
    ENDIF
    IF (nHandle > -1)
        FCLOSE(nHandle)
    ENDIF
RETURN(aRetVal)
* ========================================================================= *
* LoadLevel  - Read in one dimension from disk, recursively processing any  *
*              arrays which may be part of that dimension.                  *
*                                                                           *
* Parameters: <ExpN1> - the handle for the file from which array is to be   *
*                       read.                                               *
*                                                                           *
* Returns   : <ExpA>  - the dimension read.                                 *
* ========================================================================= *
STATIC FUNCTION LoadLevel(nHandle)
    LOCAL aRetVal, cBuffer, lFinished, nElems, nCntr, nBytes
    LOCAL cType, nLength
    aRetVal    := {}
    cBuffer    := SPACE(NELEMLEN)
    lFinished  := !ReadFile(nHandle, @cBuffer, NELEMLEN)
    IF !lFinished
        nElems := VAL(cBuffer)
        FOR nCntr := 1 TO nElems
            lFinished := !ReadFile(nHandle, @cBuffer, TYPELEN)
            IF !lFinished
                cType     := cBuffer
                IF (cType != "A")
                    lFinished := !ReadFile(nHandle, @cBuffer, VALUELEN)
                    IF !lFinished
                        nLength   := VAL(cBuffer)
                        lFinished := !ReadFile(nHandle, @cBuffer, nLength)
                    ENDIF
                ENDIF
                DO CASE
                    CASE lFinished
                        * Can't read file, so do nothing.
                    CASE cType == "A"
                        * Array
                        AADD(aRetVal, LoadLevel(nHandle))
                    CASE (cType $ "BCMO")
                        * Block, Character, Memo, Object
                        AADD(aRetVal, cBuffer)
                    CASE cType == "D"
                        * Date
                        AADD(aRetVal, STOD(cBuffer))
                    CASE cType == "L"
                        * Logical
                        AADD(aRetVal, (cBuffer == "T"))
                    CASE cType == "N"
                        * Number
                        AADD(aRetVal, VAL(cBuffer))
                ENDCASE
            ENDIF
            IF lFinished                // Can't read file, so
                nCntr := nElems + 1     // force exit from loop
            ENDIF
        NEXT
    ENDIF
RETURN(aRetVal)
* ========================================================================= *
* ReadFile   - Read specified number of bytes into a buffer from a file.    *
*                                                                           *
* Parameters: <ExpN1> - the handle of the file to be read.                  *
*             <ExpC1> - the buffer into which the data is to be placed      *
*                       (specified by reference).                           *
*             <ExpN2> - the number of bytes to be read.                     *
*                                                                           *
* Returns   : <ExpL>  - was the read successful?                            *
* ========================================================================= *
STATIC FUNCTION ReadFile(nHandle, cBuffer, pBytes)
    cBuffer := SPACE(pBytes)
RETURN((FREAD(nHandle, @cBuffer, pBytes) == pBytes))
